import { TypeInfo, visit, visitWithTypeInfo, getNullableType, getNamedType, isListType, isNullableType, parse as parse$1, GraphQLError, Source, stripIgnoredCharacters, validate as validate$1, buildSchema } from 'graphql';
import { memoize, memoize2 } from '@mo36924/memoize';
import LRU from 'lru-cache';
import { readFileSync, watch, watchFile, unwatchFile, promises } from 'fs';
import { graphql } from '@mo36924/config';
import { schema } from '@mo36924/graphql-schema';

const getTypescriptType = (ts, checker, namedType) => {
  const symbol = checker.getGlobalSymbol("GraphQL", ts.SymbolFlags.Namespace);

  const _exports = symbol && checker.getExportsOfSymbol(symbol);

  const typeSymbol = _exports && checker.getSymbol(_exports, namedType.name, ts.SymbolFlags.Type);

  const type = typeSymbol && checker.getDeclaredTypeOfSymbol(typeSymbol) || checker.getUnknownType();
  return type;
};

const getArgs = (ts, schema, node, checker, documentNode) => {
  let operationDefinition;

  for (const definition of documentNode.definitions) {
    if (definition.kind === "OperationDefinition") {
      if (operationDefinition === undefined) {
        operationDefinition = definition;
      } else {
        return undefined;
      }
    }
  }

  if (operationDefinition === undefined) {
    return undefined;
  }

  const {
    createTupleType,
    getLiteralType,
    createSymbolTable,
    createSymbol,
    createAnonymousType,
    emptyArray,
    getIntersectionType,
    createSyntheticExpression,
    createArrayType,
    getUnionType,
    getNullType
  } = checker;

  const createPropertySymbolWithType = (name, type, readonly) => {
    const symbol = createSymbol(ts.SymbolFlags.Property, name, readonly ? 8 : undefined);
    symbol.type = type;
    return symbol;
  };

  const typeInfo = new TypeInfo(schema);
  const values = [];
  const variables = [];
  const symbols = [];
  const symbolsMap = new Map();
  visit(operationDefinition, visitWithTypeInfo(typeInfo, {
    VariableDefinition(node) {
      const variableName = node.variable.name.value;
      const inputType = typeInfo.getInputType();
      const nullableType = getNullableType(inputType);
      const namedType = getNamedType(nullableType);
      let type = getTypescriptType(ts, checker, namedType);

      if (isListType(nullableType)) {
        type = createArrayType(type);
      }

      if (isNullableType(inputType)) {
        type = getUnionType([type, getNullType()]);
      }

      const symbol = createPropertySymbolWithType(variableName, type);
      values.push(type);
      variables.push(symbol);
    },

    Field: {
      enter(node, _key, parent) {
        if (node.selectionSet) {
          symbolsMap.set(node.selectionSet.selections, []);
          return;
        }

        const parentSymbols = symbolsMap.get(parent) || symbols;
        const fieldName = (node.alias ?? node.name).value;
        const outputType = typeInfo.getType();
        const namedType = getNamedType(outputType);
        let type = getTypescriptType(ts, checker, namedType);

        if (isNullableType(outputType)) {
          type = getUnionType([type, getNullType()]);
        }

        const symbol = createPropertySymbolWithType(fieldName, type);
        parentSymbols.push(symbol);
        return false;
      },

      leave(node, _key, parent) {
        const parentSymbols = symbolsMap.get(parent) || symbols;
        const fieldName = (node.alias ?? node.name).value;
        const outputType = typeInfo.getType();
        const nullableType = getNullableType(outputType);
        const selectionSymbols = symbolsMap.get(node.selectionSet.selections);
        const selectionSymbolTable = createSymbolTable(selectionSymbols);
        let type = createAnonymousType(undefined, selectionSymbolTable, emptyArray, emptyArray, undefined, undefined);

        if (isListType(nullableType)) {
          type = createArrayType(type);
        }

        if (isNullableType(outputType)) {
          type = getUnionType([type, getNullType()]);
        }

        const symbol = createPropertySymbolWithType(fieldName, type);
        parentSymbols.push(symbol);
      }

    }
  }));
  const valuesSymbol = createPropertySymbolWithType("$values", createTupleType(values, undefined));
  const variablesSymbol = createPropertySymbolWithType("$variables", createAnonymousType(undefined, createSymbolTable(variables), emptyArray, emptyArray, undefined, undefined));
  const resultSymbol = createPropertySymbolWithType("$result", createAnonymousType(undefined, createSymbolTable(symbols), emptyArray, emptyArray, undefined, undefined));
  const template = node.template;
  const args = [];

  if (ts.isNoSubstitutionTemplateLiteral(template)) {
    const arrayType = createTupleType([getLiteralType(template.text)], undefined, true);
    const rawSymbol = createPropertySymbolWithType("raw", createTupleType([getLiteralType(template.rawText)], undefined, true), true);
    const symbolTable = createSymbolTable([rawSymbol, valuesSymbol, variablesSymbol, resultSymbol]);
    const objectType = createAnonymousType(undefined, symbolTable, emptyArray, emptyArray, undefined, undefined);
    const stringsType = getIntersectionType([arrayType, objectType]);
    args.push(createSyntheticExpression(template, stringsType));
  } else {
    const texts = [getLiteralType(template.head.text)];
    const rawTexts = [getLiteralType(template.head.rawText)];

    for (const span of template.templateSpans) {
      texts.push(getLiteralType(span.literal.text));
      rawTexts.push(getLiteralType(span.literal.rawText));
      args.push(span.expression);
    }

    const arrayType = createTupleType(texts, undefined, true);
    const rawSymbol = createPropertySymbolWithType("raw", createTupleType(rawTexts, undefined, true), true);
    const symbolTable = createSymbolTable([rawSymbol, valuesSymbol, variablesSymbol, resultSymbol]);
    const objectType = createAnonymousType(undefined, symbolTable, emptyArray, emptyArray, undefined, undefined);
    const stringsType = getIntersectionType([arrayType, objectType]);
    args.unshift(createSyntheticExpression(template, stringsType));
  }

  return args;
};

const isGraphqlTag = tagName => {
  switch (tagName) {
    case "gql":
    case "useQuery":
    case "useMutation":
      return true;

    default:
      return false;
  }
};

const parse = memoize(source => {
  try {
    return parse$1(source);
  } catch (err) {
    if (err instanceof GraphQLError) {
      return err;
    }

    return new GraphQLError(String(err));
  }
}, new WeakMap());

const source = memoize(query => new Source(query), new LRU(1000));

const minify = memoize(source$1 => {
  try {
    return source(stripIgnoredCharacters(source$1));
  } catch {
    return source$1;
  }
}, new WeakMap());

const validate = memoize2((schema, documentNode) => validate$1(schema, documentNode), new WeakMap());

const query = (ts, schema, node) => {
  const template = node.template;
  let query = "";
  let args = "";

  if (ts.isNoSubstitutionTemplateLiteral(template)) {
    // 2 ``
    const templateWidth = template.getWidth() - 2;
    query = template.text.padStart(templateWidth);
  } else {
    const head = template.head;
    const templateSpans = template.templateSpans; // 3 `...${

    const templateWidth = head.getWidth() - 3;
    query = head.text.padStart(templateWidth);

    for (let i = 0, len = templateSpans.length; i < len; i++) {
      const templateSpan = templateSpans[i];
      const templateSpanWidth = templateSpan.getFullWidth();
      const literal = templateSpan.literal;
      const literalWidth = literal.getWidth();
      const expressionWidth = templateSpanWidth - literalWidth;
      const variableName = `$_${i}`;
      const variable = variableName.padStart(expressionWidth + 2).padEnd(expressionWidth + 3);
      const templateWidth = literalWidth - (ts.isTemplateTail(literal) ? 2 : 3);
      const template = literal.text.padStart(templateWidth);
      query += variable + template;
      args += variableName + ":Unknown";
    }
  }

  const tagName = node.tag.getText();
  let offset = template.getStart() + 1;
  query = query.replace(/\n|\r/g, " ");

  if (tagName === "useQuery") {
    if (args.length) {
      query = `query(${args}){${query}}`;
      offset -= args.length + 8;
    } else {
      query = `{${query}}`;
      offset -= 1;
    }
  } else if (tagName === "useMutation") {
    if (args.length) {
      query = `mutation(${args}){${query}}`;
      offset -= args.length + 11;
    } else {
      query = `mutation{${query}}`;
      offset -= 9;
    }
  }

  if (tagName === "useQuery" || tagName === "useMutation") {
    const documentNode = parse(minify(source(query)));

    if (!(documentNode instanceof Error)) {
      const errors = validate(schema, documentNode);

      for (const error of errors) {
        const match = error.message.match(/^Variable ".*?" of type "Unknown" used in position expecting type "(.*?)"\.$/);

        if (match) {
          query = query.replace("Unknown", match[1]);
          offset += 7 - match[1].length;
        }
      }
    }
  }

  return {
    query,
    offset
  };
};

const factory = schema => (ts, node, checker) => {
  if (!ts.isIdentifier(node.tag) || !isGraphqlTag(node.tag.text)) {
    return;
  }

  const documentNode = parse(source(query(ts, schema, node).query));

  if (documentNode instanceof Error) {
    return;
  }

  const errors = validate(schema, documentNode);

  if (errors.length) {
    return;
  }

  const args = getArgs(ts, schema, node, checker, documentNode);
  return args;
};

const {
  readFile
} = promises;

const buildSchemaWithUnknownType = gql => buildSchema(`scalar Unknown\n${schema(gql)}`);

const getSchema = async () => {
  const gql = await readFile(graphql, "utf8");
  const schema = buildSchemaWithUnknownType(gql);
  return schema;
};
const getSchemaSync = () => {
  const gql = readFileSync(graphql, "utf8");
  const schema = buildSchemaWithUnknownType(gql);
  return schema;
};

let _gql;

let _schema;

const updateSchemaAsync = async () => {
  const gql = await readFile(graphql, "utf8");

  if (_gql === gql) {
    return;
  }

  const schema = buildSchemaWithUnknownType(gql);
  _gql = gql;
  _schema = schema;
  return schema;
};

const updateSchemaSync = () => {
  try {
    const gql = readFileSync(graphql, "utf8");
    const schema = buildSchemaWithUnknownType(gql);
    _gql = gql;
    _schema = schema;
  } catch {
    _gql = "";
    _schema = buildSchemaWithUnknownType(_gql);
  }

  return _schema;
};

const watchSchema = callback => {
  try {
    const watcher = watch(graphql, async () => {
      try {
        const schema = await updateSchemaAsync();
        schema && callback(schema);
      } catch {
        watcher.close();
        watchSchema(callback);
      }
    });
  } catch {
    watchFile(graphql, async stat => {
      try {
        if (stat.isFile()) {
          const schema = await updateSchemaAsync();
          schema && callback(schema);
          unwatchFile(graphql);
          watchSchema(callback);
        }
      } catch {}
    });
  }

  _schema ?? (_schema = updateSchemaSync());
  return _schema;
};

export { factory, getArgs, getSchema, getSchemaSync, getTypescriptType, isGraphqlTag, minify, parse, query, source, validate, watchSchema };
//# sourceMappingURL=index.mjs.map
